#include "Eigen/Dense"
#include <iostream>
#include <cmath>

#include "dco.hpp"

template<typename TA, typename TP>
TA f(const TA &p, const TP &x) { return pow(p*x,2); }

template<typename TA, typename TP, int M>
Eigen::Matrix<TA,M,1> F(const TA &p, const Eigen::Matrix<TP,M,1> &x, const Eigen::Matrix<TP,M,1> &y) {
  auto m=x.size();
  Eigen::Matrix<TA,M,1> r(m);
  for (auto i=0;i<m;i++) r(i)=f(p,x(i))-y(i);
  return r;
}

template<typename T, int M>
Eigen::Matrix<T,M,1> dFdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  using DCO_T=typename dco::gt1s<T>::type;
  DCO_T p_t=p;
  dco::derivative(p_t)=1;
  Eigen::Matrix<DCO_T,M,1> r_t=F(p_t,x,y);
  auto m=x.size();
  Eigen::Matrix<T,M,1> drdp(m);
  for (auto i=0;i<m;i++) drdp(i)=dco::derivative(r_t(i));
  return drdp;
}

template<typename TA, typename TP, int M>
TA E(const TA &p, const Eigen::Matrix<TP,M,1> &x, const Eigen::Matrix<TP,M,1> &y) {
  return F(p,x,y).squaredNorm();
}

template<typename TA, typename TP, int M>
TA dEdp(const TA &p, const Eigen::Matrix<TP,M,1> &x, const Eigen::Matrix<TP,M,1> &y) {
  using DCO_T=typename dco::gt1s<TA>::type;
  DCO_T p_t=p;
  dco::derivative(p_t)=1;
  return dco::derivative(E(p_t,x,y));
}

template<typename T, int M>
T ddEdpp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  using DCO_T=typename dco::gt1s<T>::type;
  DCO_T p_t=p;
  dco::derivative(p_t)=1;
  return dco::derivative(dEdp(p_t,x,y));
}

template<typename T, int M>
T Householder(Eigen::Matrix<T,M,1> drdp, Eigen::Matrix<T,M,1> r) {
  using VT=Eigen::Matrix<T,M,1>; 
  auto m=drdp.size();
  VT v=drdp+drdp(0)/fabs(drdp(0))*drdp.norm()*VT::Unit(m,0);
  drdp-=2*v.dot(drdp)/v.dot(v)*v;
  r-=2*v.dot(r)/v.dot(v)*v;
  return r(0)/drdp(0);
}

template<typename T, int M>
T Regression(const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) { 
  T p=1;
  while (fabs(dEdp(p,x,y))>eps) {
    Eigen::Matrix<T,M,1> r=F(p,x,y), drdp=dFdp(p,x,y);
    p-=Householder(drdp,r);
  }
  return p;
}

int main(int argc, char* argv[]) {
  assert(argc==2); auto M=std::stoi(argv[1]);
  using T=double; 
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  VT x=VT::Random(M), y=VT::Random(M); 
  x=x.cwiseProduct(x);
  std::cout << "x^T=" << x.transpose() << std::endl
            << "y^T=" << y.transpose() << std::endl;
  T p=Regression(x,y,1e-5);
  std::cout << "p=" << p << std::endl;
  std::cout << "E'=" << dEdp(p,x,y) << std::endl;
  std::cout << "E''=" << ddEdpp(p,x,y) << std::endl;
  return 0;
}
