#include "Eigen/Dense"
#include <iostream>
#include <limits>
#include <cmath>

template<typename T, int M>
Eigen::Matrix<T,M,1> F(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  return p*p*x.cwiseProduct(x)-y;
}

template<typename T, int M>
Eigen::Matrix<T,M,1> dFdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T dp=(p==0) ? sqrt(std::numeric_limits<T>::epsilon())
              : sqrt(std::numeric_limits<T>::epsilon())*fabs(p);
  return (F(p+dp,x,y)-F(p,x,y))/dp;
}

template<typename T, int M>
T E(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  return F(p,x,y).squaredNorm();
}

template<typename T, int M>
T dEdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T dp=(p==0) ? sqrt(sqrt(std::numeric_limits<T>::epsilon()))
              : sqrt(sqrt(std::numeric_limits<T>::epsilon()))*fabs(p);
  return (E(p+dp,x,y)-E(p-dp,x,y))/(2*dp);
}

template<typename T, int M>
T ddEdpp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  T dp=(p==0) ? sqrt(sqrt(std::numeric_limits<T>::epsilon()))
              : sqrt(sqrt(std::numeric_limits<T>::epsilon()))*fabs(p);
  return (dEdp(p+dp,x,y)-dEdp(p-dp,x,y))/(2*dp);
}

template<typename T, int M>
T Householder(Eigen::Matrix<T,M,1> drdp, Eigen::Matrix<T,M,1> r) {
  using VT=Eigen::Matrix<T,M,1>; 
  auto m=drdp.size();
  VT v=drdp+drdp(0)/fabs(drdp(0))*drdp.norm()*VT::Unit(m,0);
  drdp-=2*v.dot(drdp)/v.dot(v)*v;
  r-=2*v.dot(r)/v.dot(v)*v;
  return r(0)/drdp(0);
}

template<typename T, int M>
T Regression(const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) { 
  T p=1;
  while (fabs(dEdp(p,x,y))>eps) {
    Eigen::Matrix<T,M,1> r=F(p,x,y), drdp=dFdp(p,x,y);
    p-=Householder(drdp,r);
  }
  return p;
}

int main(int argc, char* argv[]) {
  assert(argc==2); auto M=std::stoi(argv[1]);
  using T=double; 
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  VT x=VT::Random(M), y=VT::Random(M); 
  x=x.cwiseProduct(x);
  std::cout << "x^T=" << x.transpose() << std::endl
            << "y^T=" << y.transpose() << std::endl;
  T p=Regression(x,y,1e-5);
  std::cout << "p=" << p << std::endl;
  std::cout << "E'=" << dEdp(p,x,y) << std::endl;
  std::cout << "E''=" << ddEdpp(p,x,y) << std::endl;
  return 0;
}
