#include "Eigen/Dense"
#include <iostream>
#include <cmath>

template<typename T>
T f(const T &p, const T &x) { return pow(p*x,2); }

template<typename T>
T dfdp(const T &p, const T &x) { return 2*p*pow(x,2); }

template<typename T>
T ddfdpp(const T&, const T &x) { return 2*pow(x,2); }

template<typename T, int M>
Eigen::Matrix<T,M,1> F(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  auto m=x.size();
  Eigen::Matrix<T,M,1> r(m);
  for (int i=0;i<m;i++) r(i)=f(p,x(i))-y(i);
  return r;
}

template<typename T, int M>
Eigen::Matrix<T,M,1> dFdp(const T &p, const Eigen::Matrix<T,M,1> &x) {
  auto m=x.size();
  Eigen::Matrix<T,M,1> drdp(m);
  for (int i=0;i<m;i++) drdp(i)=dfdp(p,x(i));
  return drdp;
}

template<typename T, int M>
T dEdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  return 4*p*(p*p*(x.cwiseProduct(x)).dot(x.cwiseProduct(x))-(x.cwiseProduct(x)).dot(y));
}

template<typename T, int M>
T ddEdpp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  return 4*(3*p*p*(x.cwiseProduct(x)).dot(x.cwiseProduct(x))-(x.cwiseProduct(x)).dot(y));
}

template<typename T, int M>
T Householder(Eigen::Matrix<T,M,1> drdp, Eigen::Matrix<T,M,1> r) {
  using VT=Eigen::Matrix<T,M,1>; 
  auto m=drdp.size();
  VT v=drdp+drdp(0)/fabs(drdp(0))*drdp.norm()*VT::Unit(m,0);
  drdp-=2*v.dot(drdp)/v.dot(v)*v;
  r-=2*v.dot(r)/v.dot(v)*v;
  return r(0)/drdp(0);
}

template<typename T, int M>
T Regression(const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) { 
  T p=1;
  while (fabs(dEdp(p,x,y))>eps) {
    Eigen::Matrix<T,M,1> r=F(p,x,y), drdp=dFdp(p,x);
    p-=Householder(drdp,r);
  }
  return p;
}

int main(int argc, char* argv[]) {
  assert(argc==2); auto M=std::stoi(argv[1]);
  using T=double; 
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  VT x=VT::Random(M), y=VT::Random(M); 
  x=x.cwiseProduct(x);
  std::cout << "x^T=" << x.transpose() << std::endl
            << "y^T=" << y.transpose() << std::endl;
  T p=Regression(x,y,1e-5);
  std::cout << "p=" << p << std::endl;
  std::cout << "E'=" << dEdp(p,x,y) << std::endl;
  std::cout << "E''=" << ddEdpp(p,x,y) << std::endl;
  return 0;
}
