#include "Eigen/Dense"
#include <iostream>
#include <cassert>
#include <fstream>

template<typename T>
T f(const T &p, const T &x) { return pow(p*x,2); }

template<typename T>
T dfdp(const T &p, const T &x) { return 2*p*pow(x,2); }

template<typename T, int M>
Eigen::Matrix<T,M,1> F(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  auto m=x.size();
  Eigen::Matrix<T,M,1> r(m);
  for (int i=0;i<m;i++) r(i)=f(p,x(i))-y(i);
  return r;
}

template<typename T, int M>
Eigen::Matrix<T,M,1> dFdp(const T &p, const Eigen::Matrix<T,M,1> &x) {
  return 2*p*x.cwiseProduct(x);
}

template<typename T, int M>
T dEdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  Eigen::Matrix<T,M,1> xx=x.cwiseProduct(x);
  return 4*p*(p*p*xx.dot(xx)-xx.dot(y));
}

template<typename T, int M>
T ddEdpp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  Eigen::Matrix<T,M,1> xx=x.cwiseProduct(x);
  return 4*(3*p*p*xx.dot(xx)-xx.dot(y));
}

template<typename T, int M>
T NormalEquation(const Eigen::Matrix<T,M,1> &drdp, const Eigen::Matrix<T,M,1> &r) {
  return drdp.dot(r)/drdp.dot(drdp);
}

template<typename T, int M>
T Regression(const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) { 
  T p=1;
  while (fabs(dEdp(p,x,y))>1e-7) {
    Eigen::Matrix<T,M,1> r=F(p,x,y), drdp=dFdp(p,x);
    p-=NormalEquation(drdp,r);
  }
  return p;
}

int main(int argc, char* argv[]) {
  assert(argc==2); auto M=std::stoi(argv[1]);
  using T=double; 
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  VT x=VT::Random(M), y=VT::Random(M); 
  x=x.cwiseProduct(x);
  std::cout << "x^T=" << x.transpose() << std::endl
            << "y^T=" << y.transpose() << std::endl;
  T p=Regression(x,y);
  std::cout << "p=" << p << std::endl;
  std::cout << "E'=" << dEdp(p,x,y) << std::endl;
  std::cout << "E''=" << ddEdpp(p,x,y) << std::endl;

  std::ofstream out("data");
  for (auto i=0;i<M;i++) out << x(i) << " " << y(i)<< std::endl;
  return 0;
}
